import os
from typing import Any, Dict

import click
from cookiecutter.main import cookiecutter
from mach import utils

DEFAULT_COOKIECUTTER = "git@github.com:labd/mach-component-cookiecutter.git"


def create_component(output_dir: str, cookiecutter_location: str):
    """Create a component using the given cookiecutter template or the default one."""
    cookiecutter_kwargs: Dict[str, Any] = {
        "output_dir": output_dir or ".",
    }

    if output_dir and os.path.exists(output_dir):
        if not click.confirm(
            f"Directory {output_dir} already exists. Do you want to overwrite?"
        ):
            return

    if not cookiecutter_location:
        cookiecutter_location = DEFAULT_COOKIECUTTER
        cloud = click.prompt(
            "Cloud environment", type=click.Choice(["aws", "azure"]), default="aws"
        )

        cookiecutter_kwargs.update(
            {
                "directory": cloud,
                "extra_context": _get_component_context(cloud),
                "no_input": True,
            }
        )

    result = cookiecutter(
        cookiecutter_location,
        **cookiecutter_kwargs,
    )

    dirname = os.path.basename(result)
    click.echo(f"New component {dirname} created 🎉")


def _get_component_context(cloud: str) -> dict:
    """Prompt user for input for the cookiecutter."""
    # We have a different default language here because the cookiecutter
    # doesnt contain examples for all languages yet.
    default_lang = "node" if cloud == "aws" else "python"

    language = click.prompt(
        "Language", type=click.Choice(["python", "node"]), default=default_lang
    )

    name = click.prompt("Name", default="example-name")
    description = click.prompt(
        "Description", default=f"{utils.humanize_str(name)} component"
    )

    dirname = click.prompt(
        "Directory name", default=f"{utils.slugify(name, sep='-')}-component"
    )

    context = {
        "language": language,
        "name": name,
        "description": description,
        "component_identifier": dirname,
        "function_name": utils.slugify(name),
        "use_public_api": 0,
        "include_graphql_server": 0,
        "use_commercetools": 0,
        "use_commercetools_api_extension": 0,
        "use_commercetools_subscription": 0,
        "use_commercetools_token_rotator": 0,
    }

    if click.confirm("Uses an HTTP endpoint?", default=True):
        context["use_public_api"] = 1
        if cloud == "azure":
            context["function_name"] = click.prompt(
                "Function name", default=utils.slugify(name)
            )
        context["include_graphql_server"] = int(
            click.confirm("Include GraphQL support?", default=True)
        )

    if click.confirm("Uses commercetools?", default=True):
        context["use_commercetools"] = 1
        context["use_commercetools_api_extension"] = int(
            click.confirm("Generate commercetools API extension?", default=True)
        )
        context["use_commercetools_subscription"] = int(
            click.confirm("Generate commercetools Subcription?", default=True)
        )
        if cloud == "aws":
            context["use_commercetools_token_rotator"] = int(
                click.confirm(
                    "Do you want use the commercetools token rotator component?",
                    default=False,
                )
            )

    if click.confirm("Use Sentry?", default=False):
        context["sentry_organization"] = click.prompt("Sentry Organization")
        context["sentry_project"] = click.prompt("Sentry Project")

    if cloud == "aws":
        context["lambda_s3_repository"] = click.prompt("Lambda repository S3 bucket")
    elif cloud == "azure":
        context.update(
            {
                "shared_resource_group": click.prompt("Shared resource group name"),
                "function_storage_account": click.prompt(
                    "Function storage account name"
                ),
                "function_container_name": click.prompt(
                    "Function container name", default="code"
                ),
            }
        )

    return context



from dataclasses import dataclass
from typing import Dict, List, Optional

from dataclasses_json import dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin

from . import fields
from .shared import ComponentAzureConfig

__all__ = [
    "ComponentConfig",
    "LocalArtifact",
]


@dataclass_json
@dataclass
class LocalArtifact(JsonSchemaMixin):
    script: str
    filename: str
    workdir: Optional[str] = fields.none()


@dataclass_json
@dataclass
class ComponentConfig(JsonSchemaMixin):
    """Component definition."""

    name: str
    source: str
    version: str
    integrations: List[str] = fields.list_()
    endpoints: Dict[str, str] = fields.dict_()
    health_check_path: Optional[str] = fields.none()

    # Azure-specific options
    azure: Optional[ComponentAzureConfig] = fields.none()

    # Development options
    branch: Optional[str] = fields.none()

    artifacts: Dict[str, LocalArtifact] = fields.dict_()

    @property
    def use_version_reference(self):
        """Indicate if the module should be referenced with the version.

        This will be mainly used for development purposes when referring
        to a local directory; versioning is not possible, but we should
        still be able to define a version in our component for the actual
        function deployment itself.
        """
        return self.source.startswith("git")



from .base import *  # NOQA
from .components import *  # NOQA
from .general_config import *  # NOQA
from .mach import *  # NOQA
from .shared import *  # NOQA
from .sites import *  # NOQA
from .values import *  # NOQA



import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional, Union

import click
from mach.templates import setup_jinja
from mach.types import MachConfig, Site


def generate_terraform(config: MachConfig, *, site: str = None):
    """Generate Terraform file from template and reformat it."""
    env = setup_jinja()
    template = env.get_template("site.tf")
    sites = _filter_sites(config.sites, site)
    for s in sites:
        site_dir = config.deployment_path / Path(s.identifier)
        site_dir.mkdir(exist_ok=True)
        output_file = site_dir / Path("site.tf")
        content = _clean_tf(
            template.render(
                config=config,
                general_config=config.general_config,
                site=s,
            )
        )
        with open(output_file, "w+") as fh:
            fh.write(content)
        click.echo(f"Generated file {output_file}")

        if config.variables_path and config.variables_encrypted:
            # We need to copy the variables file for Terraform to access it
            shutil.copy(config.variables_path, f"{site_dir}/variables.yml")
            click.echo("Copied variables file")

        run_terraform("fmt", cwd=site_dir)


def _clean_tf(content: str) -> str:
    """Clean the Terraform file.

    The pyhcl (used in testing for example) doesn't like empty objects with newlines
    for example. Let's get rid of those.
    """
    return re.sub(r"\{(\s*)\}", "{}", content)


def init_terraform(config: MachConfig, *, site: str = None):
    """Terraform init for all generated sites."""
    sites = _filter_sites(config.sites, site)
    for s in sites:
        site_dir = config.deployment_path / Path(s.identifier)
        if not site_dir.is_dir():
            click.echo(f"Could not find site directory {site_dir}")
            continue

        click.echo(f"Terraform init for {site_dir.name}")
        run_terraform("init", site_dir)


def plan_terraform(
    config: MachConfig,
    *,
    site: str = None,
    components: List[str] = [],
    with_sp_login: bool = False,
    reuse=False,
    destroy=False,
):
    """Terraform init and plan for all generated sites."""
    sites = _filter_sites(config.sites, site)
    for s in sites:
        site_dir = config.deployment_path / Path(s.identifier)
        if not site_dir.is_dir():
            click.echo(f"Could not find site directory {site_dir}")
            continue

        click.echo(f"Terraform plan for {site_dir.name}")

        if not reuse:
            run_terraform("init", site_dir)

        if with_sp_login:
            azure_sp_login()

        cmd = ["plan"]

        if destroy:
            cmd += ["-destroy"]

        for component in components:
            cmd.append(f"-target=module.{component}")

        run_terraform(cmd, site_dir)


def apply_terraform(
    config: MachConfig,
    *,
    site: str = None,
    components: List[str] = [],
    with_sp_login: bool = False,
    auto_approve: bool = False,
    reuse=False,
    destroy=False,
):
    """Terraform apply for all generated sites."""
    sites = _filter_sites(config.sites, site)
    for s in sites:
        site_dir = config.deployment_path / Path(s.identifier)
        if not site_dir.is_dir():
            click.echo(f"Could not find site directory {site_dir}")
            continue

        click.echo(f"Applying Terraform for {s.identifier}")

        if not reuse:
            run_terraform("init", site_dir)

        if with_sp_login:
            azure_sp_login()

        cmd = ["apply"]

        if destroy:
            cmd += ["-destroy"]

        if auto_approve:
            cmd += ["-auto-approve"]

        for component in components:
            cmd.append(f"-target=module.{component}")
        run_terraform(cmd, site_dir)


def azure_sp_login():
    """Login the service principal with az cli."""
    p = subprocess.run(
        [
            "az",
            "login",
            "--service-principal",
            "-u",
            os.environ["ARM_CLIENT_ID"],
            "-p",
            os.environ["ARM_CLIENT_SECRET"],
            "--tenant",
            os.environ["ARM_TENANT_ID"],
        ],
        stdout=sys.stdout,
        stderr=sys.stderr,
    )
    p.check_returncode()


def run_terraform(command: Union[List[str], str], cwd):
    """Run any Terraform command."""
    if isinstance(command, str):
        command = [command]
    p = subprocess.run(
        ["terraform", *command], cwd=cwd, stdout=sys.stdout, stderr=sys.stderr
    )
    p.check_returncode()


def _filter_sites(sites: List[Site], site_identifier: Optional[str]) -> List[Site]:
    if not site_identifier:
        return sites

    return [s for s in sites if s.identifier == site_identifier]



import os

import click
import yaml
from mach import types
from mach.__version__ import __version__


def create_configuration(output_file: str):
    if os.path.exists(output_file):
        if not click.confirm(
            f"File {output_file} already exists. Do you want to overwrite?"
        ):
            return

    config = _create_config()
    data = config.to_dict()
    data = _clean_config_dump(data)
    content = yaml.dump(data, indent=2, explicit_start=True, sort_keys=False)
    with open(output_file, "w") as f:
        f.write(content)


def _create_config() -> types.MachConfig:  # noqa: C901
    environment = click.prompt("Environment", "test")
    cloud = click.prompt(
        "Cloud environment", type=click.Choice(["aws", "azure"]), default="aws"
    )
    site_id = click.prompt("Site identifier")
    use_commercetools = click.confirm("Use commercetools?", default=True)
    ct_project = ""
    if use_commercetools:
        ct_project = click.prompt("commercetools project name", default=site_id)
    use_sentry = click.confirm("Use Sentry?", default=False)
    use_contentful = click.confirm("Use Contentful?", default=False)
    use_amplience = click.confirm("Use Amplience?", default=False)

    integrations = []
    if use_commercetools:
        integrations.append("commercetools")
    if use_sentry:
        integrations.append("sentry")
    if use_contentful:
        integrations.append("contentful")
    if use_amplience:
        integrations.append("amplience")
    # If we do have integrations, add the default (cloud) integration here as well
    if integrations:
        integrations = [cloud] + integrations

    if cloud == "aws":
        tf_config = types.TerraformConfig(
            aws_remote_state=types.AWSTFState(
                bucket="<your bucket>",
                key_prefix="mach",
                region="eu-central-1",
            )
        )
    else:
        tf_config = types.TerraformConfig(
            azure_remote_state=types.AzureTFState(
                resource_group="<your-resource-group>",
                storage_account="<your-storage-account>",
                container_name="<your-container-name>",
                state_folder=environment,
            )
        )

    general_config_kwargs = dict(
        environment=environment,
        terraform_config=tf_config,
        cloud=cloud,
    )

    if cloud == "azure":
        general_config_kwargs["azure"] = types.AzureConfig(
            tenant_id="<your-tenant-id>",
            subscription_id="<your-subscription-id>",
            region="westeurope",
        )

    if use_sentry:
        general_config_kwargs["sentry"] = types.SentryConfig(
            auth_token="<your-auth-token>",
            project="<your-project>",
            organization="<your-organization>",
        )

    if use_contentful:
        general_config_kwargs["contentful"] = types.ContentfulConfig(
            cma_token="<your-cma-token>",
            organization_id="<your-organization-id>",
        )

    if use_amplience:
        general_config_kwargs["amplience"] = types.AmplienceConfig(
            client_id="<your-client-id>",
            client_secret="<your-client-secret>",
        )

    site = types.Site(
        identifier=site_id,
        components=[
            types.Component(
                name="your-component",
                variables={"FOO_VAR": "my-value"},
                secrets={"MY_SECRET": "secretvalue"},
            )
        ],
    )

    if cloud == "aws":
        site.aws = types.SiteAWSSettings(
            account_id=123456789,
            region="eu-central-1",
        )

    if use_commercetools:
        site.commercetools = types.CommercetoolsSettings(
            project_key=ct_project,
            client_id="<client-id>",
            client_secret="<client-secret>",
            scopes=(
                f"manage_api_clients:{ct_project} "
                f"manage_project:{ct_project} "
                f"view_api_clients:{ct_project}"
            ),
            project_settings=types.CommercetoolsProjectSettings(
                # TODO: Improve this by letting user select one or more countries
                # and generate/guess the correct languages and currencies that probably
                # need to be applied for that project
                languages=["nl-NL"],
                countries=["NL"],
                currencies=["EUR"],
            ),
        )

    component_config = types.ComponentConfig(
        name="your-component",
        source="git::https://github.com/<username>/<your-component>.git//terraform",
        version="0.1.0",
        integrations=integrations,
    )
    if cloud == "azure":
        component_config.azure = types.ComponentAzureConfig(
            short_name="yourcomp",
        )

    return types.MachConfig(
        mach_composer=types.MachComposerConfig(version=__version__),
        general_config=types.GlobalConfig(
            **general_config_kwargs,
        ),
        sites=[site],
        components=[component_config],
    )


def _clean_config_dump(data: dict) -> dict:
    """Perform cleanup on the dump.

    TODO: These are actions that should be performed in the Marshmallow schema.
    """
    data.pop("output_path", None)
    data.pop("file", None)
    data.pop("file_encrypted", None)

    data.pop("variables", None)
    data.pop("variables_path", None)
    data.pop("variables_encrypted", None)

    return data



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional

from dataclasses_json import config, dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin

from . import fields
from .components import ComponentConfig
from .general_config import GlobalConfig
from .mach import MachComposerConfig
from .sites import Site

__all__ = ["MachConfig"]


@dataclass_json
@dataclass
class MachConfig(JsonSchemaMixin):
    """Main MACH configuration object."""

    mach_composer: MachComposerConfig
    general_config: GlobalConfig = field(metadata=config(field_name="global"))
    sites: List[Site]
    components: List[ComponentConfig] = fields.list_()

    # Items that are not used in the configuration itself by set by the parser
    output_path: str = "deployments"
    file: Optional[str] = fields.none()
    # Indicates that the config file is SOPS encrypted
    file_encrypted: bool = False

    variables: Dict[str, Any] = fields.dict_()
    variables_path: str = fields.none()
    variables_encrypted: bool = False

    @property
    def deployment_path(self) -> Path:
        return Path(self.output_path)

    def deployment_path_for(self, site: Site):
        return self.deployment_path / Path(site.identifier)

    def get_component(self, name: str) -> Optional[ComponentConfig]:
        for comp in self.components:
            if comp.name == name:
                return comp
        return None



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from dataclasses_json import config, dataclass_json
from dataclasses_jsonschema import JsonSchemaMixin
from mach import utils
from marshmallow import ValidationError

from . import fields
from .general_config import (
    AmplienceConfig,
    AzureConfig,
    ContentfulConfig,
    FrontdoorSettings,
    SentryConfig,
)
from .shared import ComponentAzureConfig, ServicePlan

TerraformVariables = Dict[str, Any]
StoreVariables = Dict[str, TerraformVariables]
LocalizedString = Dict[str, str]
Tags = Dict[str, str]

if TYPE_CHECKING:
    from .components import ComponentConfig

__all__ = [
    "AlertGroup",
    "ApolloFederationSettings",
    "AWSProvider",
    "Endpoint",
    "CommercetoolsStore",
    "CommercetoolsChannel",
    "CommercetoolsTax",
    "CommercetoolsTaxCategory",
    "CommercetoolsProjectSettings",
    "CommercetoolsFrontendSettings",
    "CommercetoolsSettings",
    "ContentfulSettings",
    "AmplienceSettings",
    "SentryDsn",
    "Component",
    "SiteAWSSettings",
    "SiteAzureSettings",
    "Site",
]


@dataclass_json
@dataclass
class AlertGroup(JsonSchemaMixin):
    """Alert group configuration."""

    name: str
    alert_emails: List[str] = fields.list_()
    webhook_url: Optional[str] = fields.none()
    logic_app: Optional[str] = fields.none()

    @property
    def logic_app_name(self) -> Optional[str]:
        return self.logic_app.split(".")[1] if self.logic_app else None

    @property
    def logic_app_resource_group(self) -> Optional[str]:
        return self.logic_app.split(".")[0] if self.logic_app else None


@dataclass_json
@dataclass
class AWSProvider(JsonSchemaMixin):
    """AWS provider configuration."""

    name: str
    region: str
    default_tags: Tags = fields.dict_()


@dataclass_json
@dataclass
class AzureEndpoint(JsonSchemaMixin):
    session_affinity_enabled: Optional[bool] = fields.default(False)
    session_affinity_ttl_seconds: Optional[int] = fields.default(0)
    waf_policy_id: Optional[str] = fields.none()
    internal_name: Optional[str] = fields.none()


@dataclass_json
@dataclass
class AWSEndpoint(JsonSchemaMixin):
    throttling_burst_limit: Optional[int] = fields.none()
    throttling_rate_limit: Optional[int] = fields.none()
    enable_cdn: Optional[bool] = fields.default(False)


@dataclass_json
@dataclass
class Endpoint(JsonSchemaMixin):
    url: str
    key: str = field(metadata=config(exclude=lambda x: True))
    zone: Optional[str] = fields.none()
    aws: Optional[AWSEndpoint] = fields.none()
    azure: Optional[AzureEndpoint] = fields.none()

    # To be set by the parser
    components: List["Component"] = fields.list_()

    @property
    def contains_defaults(self):
        """Indicate if this endpoint contains just default values.

        Other then the `url` attribute.
        If only defaults, we can serialize the endpoints by just
        rendering the url, not the entire object.

        At this moment, we don't have any additional options, so it's always default.
        This can be extended in the future.
        """
        return True

    @property
    def subdomain(self) -> str:
        if not self.url:
            return ""

        return utils.subdomain_from_url(self.url)

    @property
    def is_root_domain(self):
        return self.url == self.zone

    def __post_init__(self):
        """Ensure endpoints have protocol stripped."""
        self.url = utils.strip_protocol(self.url)

        if not self.zone and self.url:
            try:
                self.zone = utils.dns_zone_from_url(self.url)
            except ValueError as e:
                raise ValidationError(f"Could not determine DNS zone: {e}")


@dataclass_json
@dataclass
class CommercetoolsStore(JsonSchemaMixin):
    """commercetools store definition."""

    key: str
    name: LocalizedString = fields.dict_()
    managed: bool = True
    languages: Optional[List[str]] = fields.list_()
    distribution_channels: Optional[List[str]] = fields.list_()
    supply_channels: Optional[List[str]] = fields.list_()

    def __post_init__(self):
        if self.managed and not self.name:
            raise ValidationError("name is required")


@dataclass_json
@dataclass
class CommercetoolsChannel(JsonSchemaMixin):
    """commercetools channel definition."""

    key: str
    roles: List[str]
    name: Optional[LocalizedString] = fields.none()
    description: Optional[LocalizedString] = fields.none()


@dataclass_json
@dataclass
class CommercetoolsTax(JsonSchemaMixin):
    """commercetools tax definition."""

    country: str
    amount: float
    name: str
    included_in_price: Optional[bool] = True


@dataclass_json
@dataclass
class CommercetoolsTaxCategory(JsonSchemaMixin):
    """commercetools tax categories definition."""

    key: str
    name: str
    rates: Optional[List[CommercetoolsTax]] = fields.list_()


@dataclass_json
@dataclass
class CommercetoolsZoneLocation(JsonSchemaMixin):
    country: str
    state: str = ""


@dataclass_json
@dataclass
class CommercetoolsZone(JsonSchemaMixin):
    name: str
    description: str = ""
    locations: List[CommercetoolsZoneLocation] = fields.list_()


@dataclass_json
@dataclass
class CommercetoolsFrontendSettings(JsonSchemaMixin):
    create_credentials: bool = fields.default(True)
    permission_scopes: List[str] = fields.default(
        [
            "create_anonymous_token",
            "manage_my_profile",
            "manage_my_orders",
            "manage_my_shopping_lists",
            "manage_my_payments",
            "view_products",
            "view_project_settings",
        ]
    )


@dataclass_json
@dataclass
class CommercetoolsProjectSettings(JsonSchemaMixin):
    # Project settings
    currencies: Optional[List[str]] = fields.none()
    languages: Optional[List[str]] = fields.none()
    countries: Optional[List[str]] = fields.none()
    messages_enabled: bool = fields.default(True)


@dataclass_json
@dataclass
class CommercetoolsSettings(JsonSchemaMixin):
    """commercetools configuration."""

    # Authentication settings
    project_key: str
    client_id: str
    client_secret: str
    scopes: str

    token_url: Optional[str] = fields.default(
        "https://auth.europe-west1.gcp.commercetools.com"
    )
    api_url: Optional[str] = fields.default(
        "https://api.europe-west1.gcp.commercetools.com"
    )

    # Set to false to not manage commercetools settings via mach composer
    project_settings: Optional[CommercetoolsProjectSettings] = fields.none()

    channels: Optional[List[CommercetoolsChannel]] = fields.list_()
    taxes: Optional[List[CommercetoolsTax]] = fields.list_()
    tax_categories: Optional[List[CommercetoolsTaxCategory]] = fields.list_()
    stores: List[CommercetoolsStore] = fields.list_()
    zones: List[CommercetoolsZone] = fields.list_()

    # Extra credentials
    frontend: CommercetoolsFrontendSettings = fields.none()

    def __post_init__(self):
        if not self.frontend:
            self.frontend = CommercetoolsFrontendSettings()

    @property
    def managed_stores(self):
        return [store for store in self.stores if store.managed]


@dataclass_json
@dataclass
class ContentfulSettings(JsonSchemaMixin):
    """Contentful settings."""

    space: str
    default_locale: str = "en-US"
    cma_token: str = ""
    organization_id: str = ""

    def merge(self, config: ContentfulConfig):
        self.cma_token = self.cma_token or config.cma_token
        self.organization_id = self.organization_id or config.organization_id


@dataclass_json
@dataclass
class AmplienceSettings(JsonSchemaMixin):
    """Amplience settings."""

    hub_id: str
    client_id: str = ""
    client_secret: str = ""

    def merge(self, config: AmplienceConfig):
        self.client_id = self.client_id or config.client_id
        self.client_secret = self.client_secret or config.client_secret


@dataclass_json
@dataclass
class SentryDsn(JsonSchemaMixin):
    """Specific sentry DSN settings."""

    dsn: Optional[str] = fields.none()
    project: Optional[str] = fields.none()
    rate_limit_window: Optional[int] = fields.none()
    rate_limit_count: Optional[int] = fields.none()

    @classmethod
    def from_config(cls, config: SentryConfig) -> "SentryDsn":
        return cls(
            dsn=config.dsn,
            project=config.project,
            rate_limit_window=config.rate_limit_window,
            rate_limit_count=config.rate_limit_count,
        )

    def merge(self, config: Union[SentryConfig, "SentryDsn"]):
        if not self.dsn:
            self.dsn = config.dsn
        if not self.project:
            self.project = config.project
        if not self.rate_limit_window:
            self.rate_limit_window = config.rate_limit_window
        if not self.rate_limit_count:
            self.rate_limit_count = config.rate_limit_count


@dataclass_json
@dataclass
class ApolloFederationSettings(JsonSchemaMixin):
    """Apollo Federation settings."""

    api_key: str
    graph: str
    graph_variant: str


@dataclass_json
@dataclass
class Component(JsonSchemaMixin):
    """Component configuration."""

    name: str
    variables: TerraformVariables = fields.dict_()
    secrets: TerraformVariables = fields.dict_()
    store_variables: StoreVariables = fields.dict_()
    store_secrets: StoreVariables = fields.dict_()
    health_check_path: Optional[str] = fields.none()
    sentry: Optional[SentryDsn] = fields.none()
    azure: Optional[ComponentAzureConfig] = fields.none()

    @property
    def definition(self) -> "ComponentConfig":
        return self._definition

    @definition.setter
    def definition(self, definition: "ComponentConfig"):
        self._definition = definition
        self.health_check_path = self.health_check_path or definition.health_check_path

    @property
    def integrations(self) -> List[str]:
        return self.definition.integrations

    @property
    def has_cloud_integration(self) -> bool:
        return "aws" in self.integrations or "azure" in self.integrations

    @property
    def endpoints(self) -> Dict[str, str]:
        return self.definition.endpoints


@dataclass_json
@dataclass
class SiteAWSSettings(JsonSchemaMixin):
    """Site-specific AWS settings."""

    account_id: int
    region: str
    deploy_role_name: Optional[str] = fields.none()
    default_tags: Tags = fields.dict_()
    extra_providers: Optional[List[AWSProvider]] = fields.list_()


@dataclass_json
@dataclass
class SiteAzureSettings(JsonSchemaMixin):
    """Site-specific Azure settings."""

    frontdoor: Optional[FrontdoorSettings] = fields.none()
    alert_group: Optional[AlertGroup] = fields.none()
    resource_group: str = ""
    tenant_id: str = ""  # Can overwrite values from AzureConfig
    subscription_id: str = ""  # Can overwrite values from AzureConfig
    region: str = ""  # Can overwrite values from AzureConfig
    service_object_ids: Dict[str, str] = fields.dict_()
    service_plans: Dict[str, ServicePlan] = fields.dict_()

    @classmethod
    def from_config(cls, config: AzureConfig):
        return cls(
            frontdoor=config.frontdoor,
            tenant_id=config.tenant_id,
            subscription_id=config.subscription_id,
            region=config.region,
            service_object_ids=config.service_object_ids,
            service_plans=config.service_plans,
        )

    def merge(self, config: AzureConfig):
        self.frontdoor = self.frontdoor or config.frontdoor
        self.tenant_id = self.tenant_id or config.tenant_id
        self.subscription_id = self.subscription_id or config.subscription_id
        self.region = self.region or config.region
        self.service_object_ids = self.service_object_ids or config.service_object_ids
        self.service_plans = {
            **(config.service_plans or {}),
            **(self.service_plans or {}),
        }


@dataclass_json
@dataclass
class Site(JsonSchemaMixin):
    """Site definition."""

    identifier: str
    endpoints: List[Endpoint] = field(
        default_factory=list,
        metadata=config(mm_field=fields.EndpointsField(), exclude=lambda x: not x),
    )
    commercetools: Optional[CommercetoolsSettings] = fields.none()
    contentful: Optional[ContentfulSettings] = fields.none()
    amplience: Optional[AmplienceSettings] = fields.none()
    apollo_federation: Optional[ApolloFederationSettings] = fields.none()
    azure: Optional[SiteAzureSettings] = fields.none()
    aws: Optional[SiteAWSSettings] = fields.none()
    components: List[Component] = fields.list_()
    sentry: Optional[SentryDsn] = fields.none()

    @property
    def cloud_components(self) -> List[Component]:
        """Return components with cloud platform integration."""
        return [c for c in self.components if c.has_cloud_integration]

    @property
    def used_endpoints(self) -> List[Endpoint]:
        """Return only the endpoints that are actually used by the components."""
        return [ep for ep in self.endpoints if ep.components]

    @property
    def has_cdn_endpoint(self) -> bool:
        """Check if there is an endpoint with a cdn enabled."""
        return any(ep.aws.enable_cdn for ep in self.used_endpoints)

    @property
    def used_custom_endpoints(self) -> List[Endpoint]:
        """Return custom endpoints that are used by the components."""
        return [ep for ep in self.used_endpoints if ep.url]

    @property
    def dns_zones(self) -> List[str]:
        return list({e.zone for e in self.used_endpoints if e.zone})

    @classmethod
    def json_schema(cls, *args, **kwargs):
        result = super().json_schema(*args, **kwargs)
        endpoints = result["Site"]["properties"]["endpoints"]
        result["Site"]["properties"]["endpoints"] = {
            "type": "object",
            "additionalProperties": endpoints["items"],
            "default": {},
        }
        return result



import os
import re
import subprocess
from dataclasses import dataclass
from email.utils import parseaddr
from typing import List, Optional

import click
from mach import exceptions

PRETTY_FMT = {
    "commit": "%H",
    "author": "%aN <%aE>",
    "date": "%ad",
    "message": "%s",
}

PRETTY_FMT_STR = "format:" + "|".join([fmt for fmt in PRETTY_FMT.values()])


class GitError(exceptions.MachError):
    pass


@dataclass
class Commit:
    id: str
    msg: str
    author: str


def commit(message: str):
    result = _run(["git", "status", "--short"])
    if not result:
        click.echo("No changes detected, won't commit anything")
        return

    _run(["git", "commit", "-m", message])


def add(file: str):
    _run(["git", "add", file])


def ensure_local(repo: str, dest: str):
    """Ensure the repository is present on the given dest."""
    reference = ""
    reference_match = re.match(r"(.*)(?:(?:@)(\w+))$", repo)
    if reference_match:
        repo, reference = reference_match.groups()

    if os.path.exists(dest):
        _run(["git", "pull"], cwd=dest)
    else:
        clone(repo, dest)

    if reference:
        try:
            _run(["git", "reset", "--hard", reference], cwd=dest)
        except GitError as e:
            raise GitError(f"Unable to swtich to reference {reference}: {e}")


def clone(repo: str, dest: str):
    _run(["git", "clone", repo, dest])


def history(dir: str, from_ref: str, *, branch: Optional[str] = "") -> List[Commit]:
    if branch:
        _run(["git", "checkout", branch], cwd=dir)

    cmd = ["git", "log", f"--pretty={PRETTY_FMT_STR}"]
    if from_ref:
        cmd.append(f"{from_ref}..{branch or ''}")

    lines = _run(cmd, cwd=dir).decode("utf-8").splitlines()
    commits = []
    for line in lines:
        commit_id, author, date, message = line.split("|", 3)
        author, _ = parseaddr(author)
        commits.append(
            Commit(
                id=_clean_commit_id(commit_id),
                author=author,
                msg=_clean_commit_msg(message),
            )
        )

    return commits


def _clean_commit_msg(msg: str) -> str:
    return msg


def _clean_commit_id(commit_id: str) -> str:
    """Get the correct commit ID for this commit.

    It will trim the short_id since mach and the components are using a
    different commit id format (7 chars long).
    """
    return commit_id[:7]


def _run(cmd: List, *args, **kwargs) -> bytes:
    kwargs["stderr"] = subprocess.STDOUT
    try:
        return subprocess.check_output(cmd, *args, **kwargs)
    except subprocess.CalledProcessError as e:
        raise GitError(e.output.decode() if e.output else str(e)) from e



import os.path
import subprocess
import sys
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from mach.types import LocalArtifact, MachConfig


def build_packages(config: "MachConfig", restrict_components=None):
    for component in config.components:
        if restrict_components and component.name not in restrict_components:
            continue

        for artifact, artifact_cfg in component.artifacts.items():
            build_artifact(artifact_cfg)


def build_artifact(artifact: "LocalArtifact"):
    run_script(artifact.script, artifact.workdir)

    if artifact.workdir:
        artifact.filename = os.path.abspath(
            os.path.join(artifact.workdir, artifact.filename)
        )
    else:
        artifact.filename = os.path.abspath(artifact.filename)

    if not os.path.exists(artifact.filename):
        raise ValueError(f"The file {artifact.filename} doesn't exist")


def run_script(script: str, workdir: Optional[str]):
    p = subprocess.run(
        script, stdout=sys.stdout, stderr=sys.stderr, shell=True, cwd=workdir
    )
    p.check_returncode()



import os
from contextlib import contextmanager

from mach.exceptions import MachError

_ignore_var_not_found = False


class EmptyVar(str):
    pass


IGNORED_EMPTY_VAR = EmptyVar("")


class VariableNotFound(MachError):
    def __init__(self, var_name: str, pool_name="variables"):
        super().__init__(f"Variable {var_name} not found in {pool_name}")


def resolve_variable(var, variables):
    try:
        return _resolve_variable(var, variables)
    except VariableNotFound:
        if _ignore_var_not_found:
            return IGNORED_EMPTY_VAR
        raise


def resolve_env_variable(var):
    var_value = os.environ.get(var, "")
    if not var_value:
        if _ignore_var_not_found:
            return IGNORED_EMPTY_VAR

        # TODO: Add possibility by enabling/disabling strict mode using an env var or CLI option
        raise VariableNotFound(var, "environment")
    return var_value


def _resolve_variable(var, variables):
    lookup, *remain = var.split(".", maxsplit=1)

    if isinstance(variables, list):
        try:
            lookup = int(lookup)
        except ValueError:
            raise VariableNotFound("List indicies needs a number to index")
    elif not isinstance(variables, dict):
        # We've reached the end-node which is just
        raise VariableNotFound(var)

    try:
        value = variables[lookup]
    except KeyError:
        raise VariableNotFound(var)

    if remain:
        try:
            return _resolve_variable(remain[0], value)
        except VariableNotFound:
            raise VariableNotFound(var)
    return value


@contextmanager
def ignore_variable_not_found():
    global _ignore_var_not_found
    _ignore_var_not_found = True
    yield
    _ignore_var_not_found = False



import re
import unicodedata
from typing import Tuple

import tldextract

PROTOCOL_RE = re.compile(r"^(http(s)?://)")


def strip_protocol(value: str) -> str:
    return PROTOCOL_RE.sub("", value)


def dns_zone_from_url(url: str) -> str:
    return domain_parts_from_url(url)[1]


def subdomain_from_url(url: str) -> str:
    return domain_parts_from_url(url)[0]


def domain_parts_from_url(url: str) -> Tuple[str, str]:
    ext = tldextract.extract(url)
    if not ext.subdomain:
        raise ValueError("Given URL is already top-level domain")

    sd_parts = ext.subdomain.split(".")
    parts = sd_parts[1:] + [ext.domain, ext.suffix]
    return sd_parts[0], ".".join(parts)


def humanize_str(value: str) -> str:
    return re.sub(r"[-_]+", " ", value).title()


def slugify(value, sep="_"):
    value = str(value)
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", sep, value)



import os


def cache_dir_for(file: str, *, create=True):
    cache_dir = os.path.join(
        os.getcwd(), ".mach", os.path.splitext(os.path.basename(file))[0]
    )
    if create:
        os.makedirs(cache_dir, exist_ok=True)
    return cache_dir



#!/usr/bin/env python
import os
import re

import pkg_resources
from setuptools import find_packages, setup

install_requires = []
with open("requirements.txt") as r:
    result = pkg_resources.parse_requirements(r)
    install_requires = [
        str(requirement) for requirement in pkg_resources.parse_requirements(r)
    ]

try:
    with open("README.md", "r") as fh:
        readme = re.sub(
            "<!-- start-no-pypi -->.*<!-- end-no-pypi -->\n",
            "",
            fh.read(),
            flags=re.M | re.S,
        )
except IOError:
    readme = ""

about = {}
with open(os.path.join("src", "mach", "__version__.py")) as f:
    exec(f.read(), about)

setup(
    name="mach-composer",
    version=about["__version__"],
    author="Lab Digital B.V.",
    author_email="info@labdigital.nl",
    url="https://github.com/labd/mach-composer",
    description="MACH composer",
    long_description=readme,
    long_description_content_type="text/markdown",
    zip_safe=False,
    install_requires=install_requires,
    extras_require={},
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    entry_points="""
        [console_scripts]
        mach=mach.commands:mach
    """,
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    project_urls={
        "Documentation": "https://docs.machcomposer.io",
        "Source": "https://github.com/labd/mach-composer",
    },
)
